# -*- coding: utf-8 -*-
"""Constant quantities of widely-accepted physical properties."""

from aguaclara.core.units import u

# NOTE: "#: <optional_description>"  required for Sphinx autodocumentation

#: Gravitational constant
#:
#: DEPRECATED: ``1 * u.gravity`` should be used instead
GRAVITY = 9.80665 * u.m / u.s ** 2

#: Density of water
#:
#: DEPRECATED: ``core.physchem.density_water()`` should be used instead
WATER_DENSITY = 1000 * u.kg / u.m ** 3

#: Kinematic viscosity of water
#:
#: DEPRECATED: ``core.physchem.viscosity_kinematic()`` should be used instead
WATER_NU = 1 * 10 ** -6 * u.m ** 2 / u.s

#: Atmospheric pressure
#:
#: DEPRECATED: ``1 * u.atm`` should be used instead
ATM_P = 1 * u.atm

#: Average kinematic viscosity of air
#:
#: DEPRECATED: viscosity of air should be determined by temperature and
#: atmospheric pressure
AIR_NU = 12 * u.mm ** 2 / u.s

#: The ratio of maximum energy dissipation rate in a round jet to the kinetic
#: energy per time required for the jet to travel a distance equal to its
#: diameter. See `round jet
#: <https://aguaclara.github.io/Textbook/Rapid_Mix/RM_Derivations.html?#round-jet>`_
#: in the AguaClara textbook for more details.
JET_ROUND_RATIO = 0.08

#: The ratio of maximum energy dissipation rate in a plane jet to the kinetic
#: energy per time required for the jet to travel a distance equal to its
#: diameter. This ratio applies to jets in the flocculator and sedimentation
#: tank. See `plane jet
#: <https://aguaclara.github.io/Textbook/Rapid_Mix/RM_Derivations.html?#plane-jet>`_
#: in the AguaClara textbook for more details.
JET_PLANE_RATIO = 0.0124

#: Vena contracta coefficient through an orifice with 90˚ bends. This is the
#: ratio of the flow area at the point of maximal contraction to the flow area
#: before the contraction.
VC_ORIFICE_RATIO = 0.63

#: Kozeny constant
K_KOZENY = 5



"""
Tests for the research package's floc_model functions
"""

import unittest
from aguaclara.core.units import u
import aguaclara.research.floc_model as fm


class QuantityTest(unittest.TestCase):

    def assertAlmostEqualQuantity(self, first, second, places=7):
        self.assertAlmostEqual(first.magnitude, second.magnitude, places)
        self.assertEqual(first.units, second.units, places)


class TestMaterials(QuantityTest):

    def test_Material(self):
        self.assertEqual(fm.Clay.name, 'Clay')
        self.assertEqual(fm.Clay.Diameter, 7 * 10**-6 * u.m)
        self.assertEqual(fm.Clay.Density, 2650 * u.kg/u.m**3)
        self.assertEqual(fm.Clay.MolecWeight, None)

    def test_Chemical(self):
        self.assertEqual(fm.PACl.name, 'PACl')
        self.assertEqual(fm.PACl.Diameter, 9 * 10 **-8 * u.m)
        self.assertEqual(fm.PACl.Density, 1138 * u.kg/u.m**3)
        self.assertEqual(fm.PACl.MolecWeight, 1.039 * u.kg/u.mol)
        self.assertEqual(fm.PACl.AluminumMPM, 13)
        self.assertEqual(fm.PACl.Precip, 'PACl')
        self.assertEqual(fm.PACl.PrecipName, 'PACl')
        self.assertEqual(fm.PACl.PrecipDiameter, 9 * 10 **-8 * u.m)
        self.assertEqual(fm.PACl.PrecipDensity, 1138 * u.kg/u.m**3)
        self.assertEqual(fm.PACl.PrecipMolecWeight, 1.039 * u.kg/u.mol)
        self.assertEqual(fm.PACl.PrecipAluminumMPM, 13)

        self.assertEqual(fm.Alum.name, 'Alum')
        self.assertEqual(fm.Alum.Diameter, 7 * 10 **-8 * u.m)
        self.assertEqual(fm.Alum.Density, 2420 * u.kg/u.m**3)
        self.assertEqual(fm.Alum.MolecWeight, 0.59921 * u.kg/u.mol)
        self.assertEqual(fm.Alum.AluminumMPM, 2)
        self.assertEqual(fm.Alum.Precip, 'AlOH3')
        self.assertEqual(fm.Alum.PrecipName, 'AlOH3')
        self.assertEqual(fm.Alum.PrecipDiameter, 7 * 10 **-8 * u.m)
        self.assertEqual(fm.Alum.PrecipDensity, 2420 * u.kg/u.m**3)
        self.assertEqual(fm.Alum.PrecipMolecWeight, 0.078 * u.kg/u.mol)
        self.assertEqual(fm.Alum.PrecipAluminumMPM, 1)

        self.assertEqual(fm.HumicAcid.name, 'Humic Acid')
        self.assertEqual(fm.HumicAcid.Diameter, 72 * 10**-9 * u.m)
        self.assertEqual(fm.HumicAcid.Density, 1780 * u.kg/u.m**3)
        self.assertEqual(fm.HumicAcid.MolecWeight, None)
        self.assertEqual(fm.HumicAcid.Precip, 'Humic Acid')
        self.assertEqual(fm.HumicAcid.AluminumMPM, None)
        self.assertEqual(fm.HumicAcid.Precip, 'Humic Acid')
        self.assertEqual(fm.HumicAcid.PrecipName, 'Humic Acid')
        self.assertEqual(fm.HumicAcid.PrecipDiameter, 72 * 10**-9 * u.m)
        self.assertEqual(fm.HumicAcid.PrecipDensity, 1780 * u.kg/u.m**3)
        self.assertEqual(fm.HumicAcid.PrecipMolecWeight, None)
        self.assertEqual(fm.HumicAcid.PrecipAluminumMPM, None)


class TestFunctions(QuantityTest):

    def test_dens_alum_nanocluster(self):
        self.assertAlmostEqualQuantity(fm.dens_alum_nanocluster(fm.PACl),
                                       384.44465833*u.kg/u.m**3)

    def test_dens_pacl_solution(self):
        self.assertAlmostEqualQuantity(fm.dens_pacl_solution(0.5*u.g/u.L, 298*u.degK),
                                       997.84564733*u.kg/u.m**3)

    def test_conc_precipitate(self):
        self.assertAlmostEqualQuantity(fm.conc_precipitate(0.5*u.g/u.L, fm.PACl),
                                       1.48005698*u.kg/u.m**3)

    def test_conc_floc(self):
        self.assertAlmostEqualQuantity(fm.conc_floc(0.5*u.g/u.L, 10*u.g/u.L, fm.PACl),
                                       11.48005698*u.kg/u.m**3)

    def test_moles_aluminum(self):
        self.assertAlmostEqualQuantity(fm.moles_aluminum(0.45*u.g/u.L),
                                       16.66666667*u.mol/u.m**3)

    def test_sep_dist_alumimum(self):
        self.assertAlmostEqualQuantity(fm.sep_dist_aluminum(0.45*u.g/u.L),
                                       4.63589330e-9 * u.m, 16)

    def test_particle_number_concentration(self):
        self.assertAlmostEqualQuantity(fm.particle_number_concentration(10*u.g/u.L, fm.Clay) / 10**5,
                                       21011709303102/u.m**3 / 10**5, 0)

    def test_sep_dist_clay(self):
        self.assertAlmostEqualQuantity(fm.sep_dist_clay(10*u.g/u.L, fm.Clay),
                                       3.62392782e-5 * u.m, 12)

    def test_num_nanoclusters(self):
        self.assertAlmostEqualQuantity(fm.num_nanoclusters(0.7*u.g/u.L, fm.PACl) / 10**17,
                                       7.95036025e17 / 10**17 / u.m**3)

    def test_frac_vol_floc_initial(self):
        self.assertAlmostEqualQuantity(fm.frac_vol_floc_initial(0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay),
                                       0.005074162217*u.dimensionless, 10)


class TestPFunctions(QuantityTest):

    def test_p(self):
        self.assertAlmostEqual(fm.p(100, 0.1), -3)

    def test_invp(self):
        self.assertAlmostEqual(fm.invp(-3, 0.1), 100)


class TestFractalFunctions(QuantityTest):

    def diam_fractal(self):
        self.assertAlmostEqualQuantity(fm.diam_fractal(fm.DIM_FRACTAL, 5e-5*u.m, 10),
                                       0.00101811321*u.m, 9)

    def test_num_coll_reqd(self):
        self.assertAlmostEqualQuantity(fm.num_coll_reqd(fm.DIM_FRACTAL, fm.Clay, 0.001*u.m),
                                       16.464387532*u.dimensionless)

    def test_sep_dist_floc(self):
        self.assertAlmostEqualQuantity(fm.sep_dist_floc(0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay, fm.DIM_FRACTAL, 0.001*u.m),
                                       0.001473672461*u.m, 9)

    def test_frac_vol_floc(self):
        self.assertAlmostEqualQuantity(fm.frac_vol_floc(0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.DIM_FRACTAL,  fm.Clay, 0.001*u.m),
                                       0.1636046786*u.dimensionless)

    def test_dens_floc_init(self):
        self.assertAlmostEqualQuantity(fm.dens_floc_init(0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay) / 10**3,
                                       2.2624536798e3 / 10**3 * u.kg/u.m**3)


class TestFlocculationModel(QuantityTest):

    def test_ratio_clay_sphere(self):
        self.assertAlmostEqualQuantity(fm.ratio_clay_sphere(1.5),
                                       1.164773953*u.dimensionless)

    def test_ratio_area_clay_total(self):
        self.assertAlmostEqualQuantity(fm.ratio_area_clay_total(10*u.g/u.L, fm.Clay, 0.0254*u.m, 1.5),
                                       0.9598770465*u.dimensionless)

    def test_gamma_coag(self):
        self.assertAlmostEqualQuantity(fm.gamma_coag(10*u.g/u.L, 0.5*u.g/u.L, fm.PACl, fm.Clay, 0.025*u.m, 1.5),
                                       0.999112596*u.dimensionless, 5)

    def test_gamma_humic_acid_to_coag(self):
        self.assertAlmostEqualQuantity(fm.gamma_humic_acid_to_coag(0.5*u.g/u.L, 1.5*u.g/u.L, fm.HumicAcid, fm.PACl),
                                       0.2024813861*u.dimensionless)

    def test_pacl_term(self):
        self.assertAlmostEqualQuantity(fm.pacl_term(0.025*u.m, 10*u.g/u.L, 0.5*u.g/u.L, 1.5*u.g/u.L, fm.HumicAcid, fm.PACl, fm.Clay, 1.5),
                                       0.7968108928*u.dimensionless)

    def test_alpha_pacl_clay(self):
        self.assertAlmostEqualQuantity(fm.alpha_pacl_clay(0.025*u.m, 10*u.g/u.L, 0.5*u.g/u.L, 1.5*u.g/u.L, fm.HumicAcid, fm.PACl, fm.Clay, 1.5),
                                       0.001414186281*u.dimensionless)

    def test_alpha_pacl_pacl(self):
        self.assertAlmostEqualQuantity(fm.alpha_pacl_pacl(0.025*u.m, 10*u.g/u.L, 0.5*u.g/u.L, 1.5*u.g/u.L, fm.HumicAcid, fm.PACl, fm.Clay, 1.5),
                                       0.6349075988*u.dimensionless)

    def test_alpha_pacl_nat_org_mat(self):
        self.assertAlmostEqualQuantity(fm.alpha_pacl_nat_org_mat(0.025*u.m, 10*u.g/u.L, 0.5*u.g/u.L, 1.5*u.g/u.L, fm.HumicAcid, fm.PACl, fm.Clay, 1.5),
                                       0.3223924016*u.dimensionless)

    def test_alpha(self):
        self.assertAlmostEqualQuantity(fm.alpha(0.025*u.m, 10*u.g/u.L, 0.5*u.g/u.L, 1.5*u.g/u.L, fm.HumicAcid, fm.PACl, fm.Clay, 1.5),
                                       0.9587141867*u.dimensionless)

    def test_pc_viscous(self):
        self.assertAlmostEqualQuantity(fm.pc_viscous(1*u.W/u.kg, 298*u.degK, 1*u.s, 0.025*u.m, 10*u.g/u.L, 0.5*u.g/u.L, 1.5*u.g/u.L, fm.HumicAcid, fm.PACl, fm.Clay, 1, 1.5),
                                       2.579165715*u.dimensionless)

    def test_dens_floc(self):
        self.assertAlmostEqualQuantity(fm.dens_floc(0.5*u.g/u.L, 10*u.g/u.L, fm.DIM_FRACTAL, 0.001*u.m, fm.PACl, fm.Clay, 298*u.degK),
                                       1036.3615768605*u.kg/u.m**3)

    def test_vel_term_floc(self):
        self.assertAlmostEqualQuantity(fm.vel_term_floc(0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay, fm.DIM_FRACTAL, 0.001*u.m, 298*u.degK),
                                       0.01276232413*u.m/u.s)

    def test_diam_floc_vel_term(self):
        self.assertAlmostEqualQuantity(fm.diam_floc_vel_term(0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay, 2.3, 0.05*u.m/u.s, 298*u.degK),
                                       0.002858806728*u.m, 9)

    def test_time_col_laminar(self):
        self.assertAlmostEqualQuantity(fm.time_col_laminar(1*u.W/u.kg, 298*u.degK, 0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay, 0.001*u.m, 0.025*u.m, fm.DIM_FRACTAL, 1.5),
                                       0.00065495327*u.s)

    def test_time_col_turbulent(self):
        self.assertAlmostEqualQuantity(fm.time_col_turbulent(1*u.W/u.kg, 0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay, 0.001*u.m, fm.DIM_FRACTAL),
                                       0.00895198559*u.s)


class TestKolmogorovAndViscous(QuantityTest):

    def test_eta_kolmogorov(self):
        self.assertAlmostEqualQuantity(fm.eta_kolmogorov(2*u.W/u.kg, 298*u.degK),
                                       2.44907189e-5*u.m)

    def test_lambda_vel(self):
        self.assertAlmostEqualQuantity(fm.lambda_vel(2*u.W/u.kg, 298*u.degK),
                                       0.0012245359*u.m)

    def test_diam_kolmogorov(self):
        self.assertAlmostEqualQuantity(fm.diam_kolmogorov(2*u.W/u.kg, 298*u.degK, 0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay, fm.DIM_FRACTAL),
                                       4.77577893e-6*u.m)

    def test_diam_vel(self):
        self.assertAlmostEqualQuantity(fm.diam_vel(2*u.W/u.kg, 298*u.degK, 0.5*u.g/u.L, 10*u.g/u.L, fm.PACl, fm.Clay, fm.DIM_FRACTAL),
                                       0.00078540208*u.m, 10)

    def test_diam_floc_max(self):
        self.assertRaisesRegex(FutureWarning, "diam_floc_max is deprecated and will be removed after Dec 1 2019. The underlying equation is under suspicion.")

    def test_ener_dis_diam_floc(self):
        self.assertRaisesRegex(FutureWarning, "ener_dis_diam_floc is deprecated and will be removed after Dec 1 2019. The underlying equation is under suspicion.")


class TestVelocityGradient(QuantityTest):

    def test_g_straight(self):
        self.assertAlmostEqualQuantity(fm.g_straight(1*u.mL/u.s, 0.025*u.m),
                                       0.43459910/u.s)

    def test_reynolds_rapid_mix(self):
        self.assertAlmostEqualQuantity(fm.reynolds_rapid_mix(1*u.mL/u.s, 0.025*u.m, 298*u.degK),
                                       56.83616083*u.dimensionless)

    def test_dean_number(self):
        self.assertAlmostEqualQuantity(fm.dean_number(1*u.mL/u.s, 0.025*u.m, 0.1*u.m, 298*u.degK),
                                       20.09461737*u.dimensionless)

    def test_g_coil(self):
        self.assertAlmostEqualQuantity(fm.g_coil(1*u.mL/u.s, 0.025*u.m, 0.1*u.m, 298*u.degK),
                                       0.45480492/u.s)

    def test_time_res_tube(self):
        self.assertAlmostEqualQuantity(fm.time_res_tube(0.025*u.m, 2*u.m, 1*u.mL/u.s),
                                       981.74770424*u.s)

    def test_g_time_res(self):
        self.assertAlmostEqualQuantity(fm.g_time_res(1*u.mL/u.s, 0.025*u.m, 0.1*u.m, 2*u.m, 298*u.degK),
                                       446.50368346*u.dimensionless)



"""Module containing global unit registry.

The `Pint package <https://pint.readthedocs.io/en/latest/>`_ supports arithmetic
of physical quantities, which are combinations of numerical values and units of
measurement. The units of a quantity come from Pint's unit registry. This
module contains a unit registry, named ``u``, based on Pint's unit registry.
It also defines a few additional units.

Frequently Used Units and Constants in ``u``
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Prefixes for Metric Units
-------------------------
* pico- =  1e-12 = p-
* nano- =  1e-9  = n-
* micro- = 1e-6  = u- = µ-
* milli- = 1e-3  = m-
* centi- = 1e-2  = c-
* deci- =  1e-1  = d-
* deca- =  1e+1  = da-
* hecto- = 1e2   = h-
* kilo- =  1e3   = k-
* mega- =  1e6   = M-
* giga- =  1e9   = G-
* tera- =  1e12  = T-

Units of Length
---------------
* ``u.meter`` = ``u.metre`` = ``u.m``
* ``u.inch``
* ``u.foot`` = ``u.ft``
* ``u.yard`` = ``u.yd``
* ``u.mile`` = ``u.mi``

Units of Mass
-------------
* ``u.gram`` = ``u.g``
* ``u.metric_ton`` = ``u.tonne`` = ``u.t``
* ``u.ounce`` = ``u.oz``
* ``u.pound`` = ``u.lb``
* ``u.ton``
* ``u.atomic_mass_unit`` = ``u.amu``

Units of Time
-------------
* ``u.second`` = ``u.sec`` = ``u.s``
* ``u.minute`` = ``u.min``
* ``u.hour`` = ``u.hr``
* ``u.day``
* ``u.week``
* ``u.year``

Units of Temperature
--------------------
* ``u.kelvin`` = ``u.degK``
* ``u.celsius`` = ``u.degC``
* ``u.fahrenheit`` = ``u.degF``

Units of Angle
--------------
* ``u.revolution`` = ``u.rev``
* ``u.radian`` = ``u.rad``
* ``u.degree`` = ``u.deg``

Units of Force
--------------
* ``u.newton`` = ``u.N``
* ``u.gram_force`` = ``u.gf``
* ``u.kilogram_force`` = ``u.kgf``
* ``u.pound_force`` = ``u.lbf``
* ``u.ton_force``
* ``u.kip``

Units of Frequency
------------------
* ``u.hertz`` = ``u.Hz``
* ``u.revolutions_per_minute`` = ``u.rpm``

Units of Money
--------------
* ``u.dollar`` = ``u.USD``
* ``u.lempira`` = ``u.HNL``

Units of Power
--------------
* ``u.watt`` = ``u.W``
* ``u.horsepower`` = ``u.hp``

Units of Pressure
-----------------
* ``u.pascal`` = ``u.Pa``
* ``u.bar``
* ``u.atmosphere`` = ``u.atm``
* ``u.torr``
* ``u.millimeter_Hg`` = ``u.mmHg``

Units of Volume
---------------
* ``u.liter`` = ``u.litre`` = ``u.L``
* ``u.cubic_centimeter`` = ``u.cc``
* ``u.gallon`` = ``u.gal``
* ``u.quart`` = ``u.qt``
* ``u.pint`` = ``u.pt``
* ``u.cup``
* ``u.fluid_ounce`` = ``u.floz``
* ``u.tablespoon`` = ``u.tbsp``
* ``u.teaspoon`` = ``tsp``

Other Units
-----------
* ``u.joule`` = ``u.J``
* ``u.mole`` = ``u.mol`` = ``u.equivalent`` = ``u.eq``
* ``u.NTU`` = 1.47 * (``u.mg`` / ``u.L``)

  * This turbidity-concentration relation applies to kaolinite clay and is
    obtained from `Coagulation behavior of polyaluminum chloride (Wei et al.,
    2015) <https://doi.org/10.1016/j.cjche.2015.02.003>`_.

Constants
---------
* ``u.gravity``
* ``u.molar_gas_constant`` = ``u.R``
* ``u.avogadro_number``
* ``u.boltzmann_constant``


Examples
^^^^^^^^

.. code-block:: python

    >>> from aguaclara.core.units import u
    >>> mass = 3 * u.kg
    >>> height = 4 * u.m
    >>> potential_energy = mass * height * u.gravity

    >>> potential_energy
    <Quantity(12.0, 'kilogram * meter * standard_gravity')>
    >>> potential_energy.magnitude
    12.0
    >>> potential_energy.units
    <Unit('kilogram * meter * standard_gravity')>

    >>> potential_energy.to(u.joule)
    <Quantity(117.6798, 'joule')>
    >>> potential_energy.to_base_units()
    <Quantity(117.6798, 'kilogram * meter ** 2 / second ** 2')>

"""

import os
import pint
import pandas as pd

# A global unit registry that can be used by any of other module.
unit_registry = pint.UnitRegistry(
    system='mks',
    autoconvert_offset_to_baseunit=True
)
u = unit_registry

# default formatting includes 4 significant digits.
# This can be overridden on a per-print basis with
# print('{:.3f}'.format(3 * ureg.m / 9)).
u.default_format = '.4g'
pd.options.display.float_format = '{:,.4g}'.format

u.load_definitions(os.path.join(os.path.dirname(__file__),
                                            "data", "unit_definitions.txt"))


def set_sig_figs(n=4):
    """Set the default number of significant figures used to print Pint,
    Pandas and NumPy value quantities.

    :param n: number of significant figures to display. Defaults to 4.
    :type n: int

    :Examples:

    >>> from aguaclara.core.units import set_sig_figs, u
    >>> h = 2.5532532522352543*u.m
    >>> e = 25532532522352543*u.m
    >>> print('h before sigfig adjustment:',h)
    h before sigfig adjustment: 2.553 meter
    >>> print('e before sigfig adjustment:',e)
    e before sigfig adjustment: 2.553e+16 meter
    >>> set_sig_figs(10)
    >>> print('h after sigfig adjustment:',h)
    h after sigfig adjustment: 2.553253252 meter
    >>> print('e after sigfig adjustment:',e)
    e after sigfig adjustment: 2.553253252e+16 meter
    """
    u.default_format = '.' + str(n) + 'g'
    pd.options.display.float_format = ('{:,.' + str(n) + '}').format



"""The entrance tank/flocculator assembly of an AguaClara water treatment plant
contains the entrance tank, chemical dose controller (CDC), linear flow orifice
meter (LFOM), and flocculator. It adds the initial dose of coagulant and
chlorine to the influent water, then causes flocs (accumulated coagulant and
primary particles) to aggregate.

Example:
    >>> from aguaclara.design.ent_floc import *
    >>> etf = EntTankFloc(q=20 * u.L / u.s, floc=Flocculator(hl=35 * u.cm))
    >>> round(etf.ent.l, 5)
    <Quantity(40.56721, 'inch')>
"""
from aguaclara.design.ent import *
from aguaclara.design.lfom import *
from aguaclara.design.floc import *


class EntTankFloc(Component):
    """Design an AguaClara plant's entrance tank/flocculator assembly.

    The designs of the LFOM, entrance tank, and flocculator in an AguaClara
    water treatment plant are interdependent. Use this class instead of the
    classes of the individual components to design all three at once.

    Design Inputs:
        - ``q (float * u.L / u.s)``: Flow rate (recommended, 
          defaults to 20L/s)
        - ``temp (float * u.degC)``: Water temperature (recommended,
          defaults to 20°C)
        - ``ent (EntranceTank)``: Entrance Tank
          (optional, see :class:`aguaclara.design.ent.EntranceTank`
          for defaults)
        - ``floc (Flocculator)``: Flocculator
          (optional, see :class:`aguaclara.design.floc.Flocculator` for 
          defaults)
        - ``lfom (LFOM)``: Linear Flow Orifice Meter
          (optional, see :class:`aguaclara.design.lfom.LFOM` for defaults)
    """
    def __init__(self, **kwargs): 
        self.ent = EntranceTank()
        self.floc = Flocculator()
        self.lfom = LFOM()    
        self.subcomponents = [self.ent, self.floc, self.lfom]

        super().__init__(**kwargs)
        super().set_subcomponents()

        self._design_ent_floc(self.floc.ent_l)

    def _design_ent_floc(self, ent_l):
        """Design the entrance tank and flocculator in tandem.

        Each subcomponent is redesigned until the expected length of the
        entrance tank (used to design the flocculator) is close enough to the
        actual length of the entrance tank (which should accomodate the
        flocculator's channel width).
        
        Args:
            - ``ent_l (float * u.m)``: The initial guess for the entrance tank's
              length, used to design the first iteration of the flocculator.
        """
        # Design the flocculator using a guess of the entrance tank's length.
        self.floc.ent_l = ent_l
        
        # Design the entrance tank using the flocculator's channel width.
        self.ent.floc_chan_w = self.floc.chan_w

        # Recalculate if the actual length of the entrance tank is not close
        # enough.
        if np.abs(self.ent.l.to(u.m) - ent_l) / self.ent.l > 0.01:
            self._design_ent_floc(self.ent.l)



#-*- coding: utf-8 -*-
"""
Data for materials that are used in the construction of AguaClara water
treatment plants.
"""
from aguaclara.core.units import u

#:
PVC_PIPE_ROUGH = 0.12 * u.mm

#:
CONCRETE_PIPE_ROUGH = 2 * u.mm
#:
CONCRETE_DENSITY = 2400 * (u.kg / u.m ** 3)
#:
CONCRETE_THICKNESS_MIN = 5 * u.cm

#:
REBAR_D = 0.5 * u.inch
#:
ACRYLIC_T = 1 * u.cm



# -*- coding: utf-8 -*-
"""Constant dimensions to allow for human access within a plant."""

from aguaclara.core.units import u

WALKWAY_W_DEFAULT = 1 * u.m
DRAIN_CHAN_WALKWAY_W = 1.2 * u.m
BASEMENT_STAIRS_W = 0.9 * u.m
ENTRANCE_STAIRS_W = 1.2 * u.m

HUMAN_W_MIN = 45 * u.cm
DRAIN_CHAN_H_MIN = 1.5 * u.m
PLANT_H_MIN = 0.1 * u.m

DOOR_W = 1 * u.m



"""``aguaclara`` package playground setup

Provide easy setup for using the ``aguaclara`` package and all of its modules,
classes, and functions.

Note: this module should not be used for development. Instead, individually
import only necessary modules.

Example:
    >>> from aguaclara.play import *
"""
import math
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.style
import matplotlib.pyplot as plt

import aguaclara.core.constants as con
from aguaclara.core import drills
import aguaclara.core.head_loss as k
import aguaclara.core.materials as mat
from aguaclara.core import physchem as pc
import aguaclara.core.pipes as pipe
from aguaclara.core.units import u
import aguaclara.core.utility as ut
import aguaclara.core.onshape_parser as par

from aguaclara.design.cdc import CDC
from aguaclara.design.component import Component
from aguaclara.design.ent_floc import EntTankFloc
from aguaclara.design.ent import EntranceTank
from aguaclara.design.filter import Filter
from aguaclara.design.floc import Flocculator
import aguaclara.design.human_access as ha
from aguaclara.design.lfom import LFOM
from aguaclara.design.plant import Plant
from aguaclara.design.sed_chan import SedimentationChannel
from aguaclara.design.sed_tank import SedimentationTank
from aguaclara.design.sed import Sedimentor

import aguaclara.research.environmental_processes_analysis as epa
import aguaclara.research.floc_model as fm
import aguaclara.research.procoda_parser as procoda_parser
import aguaclara.research.peristaltic_pump as peristaltic_pump
import aguaclara.research.stock_qc as stock_qc

import aguaclara as ac

def set_sig_figs(n=4):
    """Set the number of significant figures used to print Pint, Pandas, and
    NumPy quantities.

    Args:
        n (int): Number of significant figures to display.
    """
    u.default_format = '.' + str(n) + 'g'
    pd.options.display.float_format = ('{:,.' + str(n) + '}').format

set_sig_figs()
matplotlib.style.use('ggplot')



# The cache decorator does not support objects nested within datastructures or other objects. It is quite limited
# because the hash function is limited. This could be revisited and we could use a serialization library like Pickle
# if this proved to be an issue (which it most likely will when we get more complex classes)

import collections
import warnings

# The cache has key=(function name , parameter serialization) and value=
__ac_cache__ = {}


def ac_cache(method):
    def _cache(*args, **kw):
        global __cache__
        param_list = [args, kw]
        params_key = tuple([method.__name__, ac_hash(param_list)])
        try:
            value = __ac_cache__[params_key]
        except KeyError:
            value = method(*args, **kw)
            __ac_cache__[params_key] = value

        return value

    # Attempt to hash any object
    def ac_hash(hashable_object):
        if is_simple_hashable(hashable_object):
            a_hash = repr(hashable_object)
        elif isinstance(hashable_object, HashableObject):
            a_hash = hashable_object.ac_hash()
        elif isinstance(hashable_object, collections.Iterable):
            a_hash = ac_hash_iterable_into_tuple(hashable_object)
        else:
            a_hash = repr(hashable_object)
            warnings.warn("Using repr() to make a hash of {}. Please consider inheriting HashableObject class as repr "
                          "will not guarantee replicable hashing and can result in bad cache returns.".format(
                repr(hashable_object)), Warning, stacklevel=2)
        return a_hash

    def ac_hash_iterable_into_tuple(hashable_object_list):
        hash_tuple = ()
        for hashable_object in hashable_object_list:
            hash_tuple += (ac_hash(hashable_object),)
        return hash_tuple

    primitive = (int, str, bool, ...)
    def is_simple_hashable(thing):
        return type(thing) in primitive

    return _cache

class HashableObject:
    def ac_hash(self):
        return tuple(sorted(self.__dict__.items()))


from aguaclara.core.units import u
from aguaclara.design.component import Component

class Filter(Component):

    sand_layer_thick = 20 *u.cm
    n_backwash_manifold_diameter = 4 * u.inch
    max_filtration_head_loss = 30 * u.cm

    @property
    def n_tanks():
        tank = math.ceil(self.q/40)
        return (tank.magnitude)*2

    @property
    def flow_rate_per_tank():
        n_tanks = n_tanks()
        return flow_rate/n_tanks

    #def filt_vel():

    #def backwash_vel():

    #def backwash_manifold_diameter():

#Design guidelines say 11 mm/s. The success of lab-scale backwashing at
# 10 mm/s suggests that this is a reasonable and conservative value
BACKWASH_VEL = 11 * u.mm / u.s

LAYER_N = 6

LAYER_VEL = 1.833 * u.mm / u.s ##VEL_FIBER_DIST_CENTER_/N_FIBER_LAYER

##Minimum thickness of each filter layer (can be increased to accomodate
# larger pipe diameters in the bottom layer)
LAYER_H_MIN = 20 * u.cm

##center to center distance for slotted pipes
MAN_CENTER_BRANCH_DIST = 10 * u.cm

##How far the branch extends into the trunk line
MAN_BRANCH_EXTENSION_L = 2 * u.cm

##The time to drain the filter box of the water above the fluidized bed
FIBER_BACKWASH_INITIATION_BOD_TIME = 3 * u.min

##Mickey suggested this value based on lab experience. This was moved to
# Expert Inputs 12/4/16 by mrf222 as a result of feedback from Monroe and
# Skyler. In the Moroceli plant, the Fi Entrance box was overflowing
# before filtration backwash. The HL of a dirty filter has therefore been
# increased from 40 to 60 cm.
HL_DIRTY = 60 * u.cm

##This is the extra head we are going to provide on top of steady state
# backwash head loss to ensure that we can fluidize the bed to initiate
# backwash.
HL_FIBER_BACKWASH_STEADY_FLOW = 20 * u.cm

##Maximum acceptable head loss through the siphon at steady state; used to
# calculate a diameter
SIPHON_HL_MAX = 35 * u.cm

##Diameter of sand drain pipe
SAND_OUTLET_D = 2 * u.inch

##Height of the barrier between the exit box and distribution box.
BARRIER_EXIT_DISTRIBUTION_H = 10 * u.cm

##Length that the siphon pipe extends up into the plant drain channel.
#Being able to shorten the stub from which the siphon discharges into the
# main plant drain channel allows for some flexibility in the hydraulic design.
SIPHON_CHANNEL_STUB_L_MIN = 20 * u.cm

ENTRANCE_PIPE_HL_MAX = 10 * u.cm

TRUNK_D_MAX = 6 * u.inch

BACKWASH_SIPHON_D_MAX = 8 * u.inch

##Purge valves on the trunk lines are angled downwards so that sediment is
# cleared more effectively. This angle allows the tees to fit on top of one
# another at the filter wall.
TRUNK_VALVES_ANGLE = 25 * u.deg

##Purge valves on the trunk lines are angled downwards so that sediment is
# cleared more effectively. This angle allows the tees to fit on top of one
# another at the filter wall.
WEIR_THICKNESS = 5 * u.cm

BRANCH_WALL_S = 5 * u.cm

GATE_VALUE_URL = "https://confluence.cornell.edu/download/attachments/173604905/Fi-Scaled-Gate-Valve-Threaded.dwg"
BALL_VALVE_URL = "https://confluence.cornell.edu/download/attachments/173604905/FiMetalBallValve.dwg"

WALL_PLANT_FLOOR_S_MIN = 10 * u.cm

INLET_WEIR_HL_MAX = 5 * u.cm

##Dimensions get too small for construction below a certain flow rate
Q_MIN = 8 * u.L / u.s

MAN_FEMCO_COUPLING_L = 6 * u.cm

##Nominal diameter of the spacer tees in the four corners of the filter
# manifold assembly.
MAN_WING_SPACER_ND = 2 * u.inch

##Length of the vertical pipe segment following the valve on the filter
# sand drain. This stub can be capped to allow the sand in the valve to
# settle, so that the valve can be closed without damage from fluidized sand.
SAND_OUTLET_PIPE_L = 20 * u.cm

#######Elevation Safety Margins

##Minimum depth in the entrance box during backwash such that there is
# standing water over the inlet.
BACKWASH_NO_SUCK_AIR_H = 20 * u.cm

##Minimum water depth over the orifices in the siphon manifold so that air
# is not entrained.
SIPHON_NO_SUCK_AIR_H = 10 * u.cm

FLUIDIZED_BED_TO_SIPHON_H = 20 * u.cm

FORWARD_NO_SUCK_AIR_H = 10 * u.cm

WEIR_FREEFALL_H = 3 * u.cm

AIR_REMOVAL_BLOCK_SUBMERGED_H = 5 * u.cm

BYPASS_SAFETY_H = 10 * u.cm

DRAIN_OUTLET_SAFETY_H = 10 * u.cm

OVERFLOW_WEIR_FREEFALL_H = 10 * u.cm

#We are going to take this pipe size for the slotted pipes as a given.
# Larger pipes may block too much flow and they are harder to install.
MAN_BRANCH_ND = 1*u.inch

MAN_BRANCH_BACKWASH_ND = 1.5*u.inch

#A slot thickness of 0.008 in or 0.2 mm is selected so that sand
# will not enter the slotted pipes.
MANIFOLD_SLOTS_W = 0.008*u.inch

BRANCH_HOLDER_ND = 2*u.inch

BACKWASH_BRANCH_HOLDER_ND = 2*u.inch

#Minimum vertical spacing between trunk line pipes going through
#the filter wall for concrete construction
TRUNK_S_MIN = 3 * u.cm

#Space between the ends of the branch receiver pipes and the walls so that
#the manifold assemblies are easy to lower into the filter boxes
# (if the branch receivers extended the entire length of the box they would
#just barely fit and it would be hard to get into place)
MAN_ASSEMBLY_S = 1 * u.cm

##Sand Properties

SAND_D_EFFECTIVE = 0.5 * u.mm

SAND_UNIFORMITY_RATIO = 1.65

DIAM_FILTER_SAND_60 = SAND_D_EFFECTIVE * SAND_UNIFORMITY_RATIO

#Porosity in a sand bed
SAND_POROSITY = 0.4

SAND_DENSITY = 2650 * u.kg / (u.m ** 3)

FLUIDIZED_RATIO = 1.3 #Bed expands 30% when fluidized



'''''
research tests.
'''''

import unittest
import aguaclara.research.environmental_processes_analysis as epa
from aguaclara.core.units import u
import numpy as np


class TestEPA(unittest.TestCase):
    '''''
    Test research's Environmental_Processes_Analysis
    '''''
    def assertAlmostEqualSequence(self, a, b, places=7):
        for elt_a, elt_b in zip(a, b):
            self.assertAlmostEqual(elt_a, elt_b, places)

    def test_Hplus_concentration_from_pH(self):
        '''''
        Test function that converts pH to molarity of H+
        '''''
        output = epa.invpH(8.25)
        self.assertEqual(output, 5.623413251903491e-09*u.mol/u.L)

        output = epa.invpH(10)
        self.assertEqual(output, 1e-10*u.mol/u.L)

    def test_E_Advective_Dispersion(self):
        output = epa.E_Advective_Dispersion(0.5, 5)
        self.assertAlmostEqual(output, 0.4774864115)

        output = epa.E_Advective_Dispersion(0, 5)
        self.assertAlmostEqual(output, 0)

        output = epa.E_Advective_Dispersion(np.array([0, 0.5, 1, 1.5, 2]), 5)
        answer = np.array([0, 0.477486411, 0.630783130, 0.418173418, 0.238743205])
        self.assertAlmostEqualSequence(output, answer)



"""Utility functions and features

This module provides functions and features for scientific calculations and
complex function inputs.

Example:
    >>> import aguaclara.core.utility as ut
    >>> ut.round_sig_figs(1234567, 3)
    1230000
"""
from aguaclara.core.units import u
import numpy as np
from math import log10, floor, ceil
import warnings
import functools


def optional_units(arg_positions, keys):
    """Wrap a function so that arguments may optionally have units.

    This should be used as a function decorator; it will not do anything
    meaningful unless it is appended with a ``@`` before another function.

    Example:
        @optional_units([0, 1], ['num', 'step'])
        def stepper(num, step=10, func=round):
            step = step.to(num.units)
            num = func(num / step) * step
            return num

    Args:
        - ``arg_positions (int list)``: Position indices of positional and
          keyword arguments with optional units
        - ``keys (str list)``: Names of positional and keyword arguments with
          optional units
    """
    # func is the function that is being decorated, and *args/**kwargs are the
    # arguments being passed to that function.
    def decorator(func):
        def wrapper(*args, **kwargs):
            args = list(args)
            for i in range(len(args)):
                if i in arg_positions:
                    args[i] *= u.dimensionless

            for key in list(kwargs.keys()):
                if key in keys:
                    kwargs[key] *= u.dimensionless

            result = func(*args, **kwargs)

            # Convert back to native Python number type, if possible
            if result.units == u.dimensionless:
                result = result.magnitude

            return result
        return wrapper
    return decorator


@optional_units([0], ['num'])
def round_sig_figs(num, figs=4):
    """Round a number to some amount of significant figures.

    Args:
        - ``num (float)``: Value to be rounded (optional units)
        - ``figs (int)``: Number of significant digits to be rounded to
          (recommended, defaults to 4)
    """
    # Prevents undefined log10(0) if num is already 0
    if num.magnitude != 0:
        decimals = figs - int(floor(log10(abs(num.magnitude)))) - 1
        num = np.round(num.magnitude, decimals) * num.units

    return num


def round_sf(num, figs=4):
    """Round a number to some amount of significant figures.

    Args:
        - ``num (float)``: Value to be rounded (optional units)
        - ``figs (int)``: Number of significant digits to be rounded to
          (recommended, defaults to 4)

    Note: This function will be deprecated after 21 Dec 2019. Use
    round_sig_figs instead.
    """
    warnings.warn(
        'round_sf will be deprecated after 21 Dec 2019. Use '
            'round_sig_figs instead.',
        FutureWarning
    )
    round_sig_figs(num, figs = figs)

@optional_units([0, 1], ['num', 'step'])
def _stepper(num, step=10, func=round):
    """Round a number to be a multiple of some step.

    Args:
        - ``num (float)``: Value to be rounded (optional units)
        - ``step (float)``: Factor to which ``num`` will be rounded (defaults
          to 10).
        - ``func (function)``: Rounding function to use (defaults to round())

    Note:
        ``step`` must have the same dimensionality as ``num``, but not
        necessarily the same units (e.g. ``num``: meters and ``step``:
        centimeters are acceptable).
    """
    step = step.to(num.units)
    num = func(num / step) * step
    return num

def round_step(num, step=10):
    """Round a number to be a multiple of some step.

    Args:
        - ``num (float)``: Value to be rounded (optional units)
        - ``step (float)``: Factor to which ``num`` will be rounded (defaults
          to 10).

    Note:
        ``step`` must have the same dimensionality as ``num``, but not
        necessarily the same units (e.g. ``num``: meters and ``step``:
        centimeters are acceptable).
    """
    return _stepper(num, step = step, func = round)

def ceil_step(num, step=10):
    """Like :func:`round_step`, but ``num`` is always rounded up."""
    return _stepper(num, step = step, func = ceil)

def floor_step(num, step=10):
    """Like :func:`round_step`, but ``num`` is always rounded down."""
    return _stepper(num, step = step, func = floor)

def stepceil_with_units(param, step, unit):
    """Round a number up to be a multiple of some step.

    Args:
        - ``param (float)``: Value to be rounded (optional units)
        - ``step (float)``: Factor to which ``param`` will be rounded
        - ``unit (Quantity)``: units of ``step``

    Note: this function will be deprecated after 21 Dec 2019. Use ceil_step
    instead.
    """
    warnings.warn(
        'stepceil_with_units will be deprecated after 21 Dec 2019. Use '
            'ceil_step instead.',
        FutureWarning
    )
    counter = 0 * unit
    while counter < param.to(unit):
        counter += step * unit
    return counter

def floor_nearest(x, array):
    """Get the nearest element of a NumPy array less than or equal to a value.

    Args:
        - ``x``: Value to compare
        - ``array (numpy.array)``: Array to search
    """
    sorted_array = np.sort(array)[::-1]  # [::-1] syntax reverses array
    if x < sorted_array[-1]:
        raise ValueError(str(x) + " is smaller than all values in the array.")
    else:
        i = np.argmax(sorted_array <= x)
        return sorted_array[i]

def ceil_nearest(x, array):
    """Get the nearest element of a NumPy array greater than or equal to a value.

    Args:
        - ``x``: Value to compare
        - ``array (numpy.array)``: Array to search
    """
    sorted_array = np.sort(array)
    if x > sorted_array[-1]:
        raise ValueError(str(x) + " is larger than all values in the array.")
    else:
        i = np.argmax(sorted_array >= x)
        return sorted_array[i]

def _minmax(*args, func=np.max):
    """Get the minuimum/maximum value of some Pint quantities with units.

    Args:
        - ``func (function)``: the min/max function being used.

    Note:
        - All quantities must have the same dimensionality, but can have
          different units.
        - The output will have the same units as the first argument.

    Example:
        >>> from aguaclara.play import *
        >>> ut.max(10 * u.m, 100 * u.cm, 32 * u.cm, 40 * u.inch, 40 * u.km)
        <Quantity(40000.0, 'meter')>
    """
    base_quantity = args[0]
    lst = []

    for arg in args:
        lst.append((arg / base_quantity))

    result = func(lst) * base_quantity
    return result

def max(*args):
    """Get the maximum value of some Pint quantities with units.

    Note:
        - All quantities must have the same dimensionality, but can have
          different units.
        - The output will have the same units as the first argument.

    Example:
        >>> from aguaclara.play import *
        >>> ut.max(10 * u.m, 100 * u.cm, 32 * u.cm, 40 * u.inch, 40 * u.km)
        <Quantity(40000.0, 'meter')>
    """
    return _minmax(*args, func = np.max)

def min(*args):
    """Like :func:`max`, but the minimum of the quantites."""
    return _minmax(*args, func = np.min)

def get_sdr(spec):
    """Get the SDR of a string ``spec`` with the form \"sdrXX\".

    Args:
        - ``spec (str)``: The specification string to be parsed."""
    if spec[:3] != "sdr":
        raise ValueError('Not a valid SDR.')
    return int(spec[3:])

def list_handler():
    """Wraps a scalar function to output a NumPy array if passed one or more inputs
    as sequences (lists, tuples or NumPy arrays). For each sequence input, this
    wrapper will recursively evaluate the function with the sequence replaced
    by each of its elements and return the results in n-dimensional NumPy array,
    where n is the number of sequence inputs.

    For a function "f" of one argument, f([x_1, ..., x_n]) would be evaluated to
    [f(x_1), ..., f(x_n)]. For a function passed multiple sequences of
    dimensions d_1, ..., d_n (from left to right), the result would be a
    d_1 x ... x d_n array.
    """
    def decorate(func):
        @functools.wraps(func) # For Sphinx documentation of decorated functions
        def wrapper(*args, **kwargs):
            """Run through the wrapped function once for each array element.
            """
            # Identify the first positional argument that is a sequence.
            # Pint units must be ignored to include sequences with units.
            argsFirstSequence = None
            for num, arg in enumerate(args):  # args is a tuple
                if isinstance(arg, u.Quantity):
                    arg = arg.to_base_units().magnitude
                if isinstance(arg, (list, tuple, np.ndarray)):
                    argsFirstSequence = num
                    break
            # Identify the first keyword argument that is a sequence.
            kwargsFirstSequence = None
            for keyword, arg in kwargs.items():  # kwargs is a dictionary
                if isinstance(arg, u.Quantity):
                    arg = arg.to_base_units().magnitude
                if isinstance(arg, (list, tuple, np.ndarray)):
                    kwargsFirstSequence = keyword
                    break

            # If there are no sequences, evaluate the function.
            if argsFirstSequence is None and kwargsFirstSequence is None:
                return func(*args, **kwargs)
            # If there are sequences, iterate through them from left to right.
            # This means beginning with positional arguments.
            elif argsFirstSequence is not None:
                result = []
                argsList = list(args)
                # For each element of the leftmost sequence, evaluate the
                # function with the sequence replaced by the single element.
                # Store the results of all the elements in result.
                for arg in argsList[argsFirstSequence]:
                    # We can safely redefine the entire list argument because
                    # the new definition remains within this namespace; it does
                    # alter the loop or penetrate further up the function.
                    argsList[argsFirstSequence] = arg
                    # This recursive call creates a multi-dimensional array if
                    # there are multiple sequence arguments.
                    result.append(wrapper(*argsList, **kwargs))
            # If there are no sequences in the positional arguments, iterate
            # through those in the keyword arguments.
            else:
                result = []
                for arg in kwargs[kwargsFirstSequence]:
                    kwargs[kwargsFirstSequence] = arg
                    result.append(wrapper(*args, **kwargs))

            if isinstance(result[0], u.Quantity):
                units = result[0].units
                return np.array([r.magnitude for r in result]) * units
            else:
                return np.array(result)

        return wrapper
    return decorate


def check_range(*args):
    """Check whether passed paramters fall within approved ranges.

    Does not return anything, but will raise an error if a parameter falls
    outside of its defined range.

    Input should be passed as an array of sequences, with each sequence
    having three elements:

        [0] is the value being checked,
        [1] is the range parameter(s) within which the value should fall, and
        [2] is the name of the parameter, for better error messages.

    If [2] is not supplied, "Input" will be appended as a generic name.

    Range requests that this function understands are listed in the
    knownChecks sequence.
    """
    knownChecks = ('>0', '>=0', '0-1', '<0', '<=0', 'int', 'boolean')
    for arg in args:
        #Converts arg to a mutable list
        arg = [*arg]
        if len(arg) == 1:
            #arg[1] details what range the parameter should fall within; if
            #len(arg) is 1 that means a validity was not specified and the
            #parameter should not have been passed in its current form
            raise TypeError("No range-validity parameter provided.")
        elif len(arg) == 2:
            #Appending 'Input" to the end allows us to give more descriptive
            #error messages that do not fail if no description was supplied.
            arg.append("Input")
        #This ensures that all whitespace is removed before checking if the
        #request is understood
        arg[1] = "".join(arg[1].lower().split())
        #This block checks that each range request is understood.
        #If the request is a compound one, it must be separated into individual
        #requests for validity comprehension
        for i in arg[1].split(","):
            if i not in knownChecks:
                raise RuntimeError("Unknown parameter validation "
                                       "request: {0}.".format(i))
        if not isinstance(arg[0], (list, tuple, np.ndarray)):
            arg[0] = [arg[0]]
        for i in arg[0]:
            if '>0' in arg[1] and i <= 0:
                raise ValueError("{1} is {0} but must be greater than "
                                 "0.".format(i, arg[2]))
            if '>=0' in arg[1] and i <0:
                raise ValueError("{1} is {0} but must be 0 or "
                                 "greater.".format(i, arg[2]))
            if '0-1' in arg[1] and not 0 <= i <= 1:
                raise ValueError("{1} is {0} but must be between 0 and "
                                 "1.".format(i, arg[2]))
            if '<0' in arg[1] and i >= 0:
                raise ValueError("{1} is {0} but must be less than "
                                 "0.".format(i, arg[2]))
            if '<=0' in arg[1] and i >0:
                raise ValueError("{1} is {0} but must be 0 or "
                                 "less.".format(i, arg[2]))
            if 'int' in arg[1] and int(i) != i:
                raise TypeError("{1} is {0} but must be a numeric "
                                "integer.".format(i, arg[2]))
            if 'boolean' in arg[1] and type(i) != bool:
                raise TypeError("{1} is {0} but must be a "
                                "boolean.".format(i, arg[2]))

def array_qtys_to_strs(lst):
    """Convert Pint quantities in a NumPy array to strings.

    Args:
        - ``lst (numpy.ndarray Quantity)``: a list of values that has a Pint
            unit attached to it
    """
    return [str(value) for value in lst]



from aguaclara.core.units import u
import aguaclara.core.utility as ut


class Stock(object):
    """A stock of material in solution, with functions for calculations
    involving flow rate and concentration. A parent class to be used in
    Variable_C_Stock and Variable_Q_Stock.
    """

    def rpm(self, vol_per_rev, Q):
        return (Q / vol_per_rev)

    def T_stock(self, V_stock, Q_stock):
        return (V_stock / Q_stock)

    def M_stock(self, V_stock, C_stock):
        return C_stock * V_stock

    def V_super_stock(self, V_stock, C_stock, C_super_stock):
        return V_stock * (C_stock / C_super_stock).to(u.dimensionless)

    def dilution_factor(self, C_stock, C_super_stock):
        return (C_stock / C_super_stock).to(u.dimensionless)


class Variable_C_Stock(Stock):
    """A flow reactor with input from a stock of material of unknown
    concentration.

    :Examples:

    >>> from aguaclara.research.stock_qc import Variable_C_Stock
    >>> from aguaclara.core.units import u
    >>> reactor = Variable_C_Stock(Q_sys = 1*u.mL/u.s, C_sys = 1.4*u.mg/u.L, Q_stock = .01*u.mL/u.s)
    >>> reactor.C_stock()
    <Quantity(140.0, 'milligram / liter')>
    """

    def __init__(self, Q_sys, C_sys, Q_stock):
        """Initialize a reactor of unknown material stock concentration.

        :param Q_sys: Flow rate of the system
        :type Q_sys: float
        :param C_sys: Concentration of the material in the system
        :type C_sys: float
        :param Q_stock: Flow rate from the stock of material
        :type Q_stock: float
        """
        self._Q_sys = Q_sys
        self._C_sys = C_sys
        self._Q_stock = Q_stock

    def Q_sys(self):
        """Return the flow rate of the system.

        :return: Flow rate of the system
        :rtype: float
        """
        return self._Q_sys

    def C_sys(self):
        """Return the concentration of the material in the system.

        :return: Concentration of the material in the system
        :rtype: float
        """
        return self._C_sys

    def Q_stock(self):
        """Return the flow rate from the stock of material.

        :return: Flow rate from the stock of material
        :rtype: float
        """
        return self._Q_stock

    def C_stock(self):
        """Return the required concentration of material in the stock given a
        reactor's desired system flow rate, system concentration, and stock
        flow rate.

        :return: Concentration of material in the stock
        :rtype: float
        """
        return self._C_sys * (self._Q_sys / self._Q_stock).to(u.dimensionless)

    @ut.list_handler()
    def rpm(self, vol_per_rev):
        """Return the pump speed required for the reactor's stock of material
        given the volume of fluid output per revolution by the stock's pump.

        :param vol_per_rev: Volume of fluid pumped per revolution (dependent on pump and tubing)
        :type vol_per_rev: float

        :return: Pump speed for the material stock, in revolutions per minute
        :rtype: float
        """
        return Stock.rpm(self, vol_per_rev, self._Q_stock).to(u.rev/u.min)

    @ut.list_handler()
    def T_stock(self, V_stock):
        """Return the amount of time at which the stock of materal will be
        depleted.

        :param V_stock: Volume of the stock of material
        :type V_stock: float

        :return: Time at which the stock will be depleted
        :rtype: float
        """
        return Stock.T_stock(self, V_stock, self._Q_stock).to(u.hr)

    @ut.list_handler()
    def M_stock(self, V_stock):
        """Return the mass of undiluted material required for the stock
        concentration.

        :param V_stock: Volume of the stock of material
        :type V_stock: float

        :return: Mass of undiluted stock material
        :rtype: float
        """
        return Stock.M_stock(self, V_stock, self.C_stock())

    @ut.list_handler()
    def V_super_stock(self, V_stock, C_super_stock):
        """Return the volume of super (more concentrated) stock that must be
        diluted for the desired stock volume and required stock concentration.

        :param V_stock: Volume of the stock of material
        :type V_stock: float
        :param C_super_stock: Concentration of the super stock
        :type C_super_stock: float

        :return: Volume of super stock to dilute
        :rtype: float
        """
        return Stock.V_super_stock(self, V_stock, self.C_stock(), C_super_stock)

    @ut.list_handler()
    def dilution_factor(self, C_super_stock):
        """Return the dilution factor of the concentration of material in the
        stock relative to the super stock.

        :param C_super_stock: Concentration of the super stock
        :type C_super_stock: float

        :return: dilution factor of stock concentration over super stock concentration (< 1)
        :rtype: float
        """
        return Stock.dilution_factor(self, self.C_stock(), C_super_stock)


class Variable_Q_Stock(Stock):
    """A flow reactor with input from a stock of material at an unknown flow
    rate.

    :Examples:

    >>> from aguaclara.research.stock_qc import Variable_Q_Stock
    >>> from aguaclara.core.units import u
    >>> reactor = Variable_Q_Stock(Q_sys = 1*u.mL/u.s, C_sys = 1.4*u.mg/u.L, C_stock = 7.6*u.mg/u.L)
    >>> round(reactor.Q_stock(), 6)
    <Quantity(0.184211, 'milliliter / second')>
    >>> pump_speed = reactor.rpm(vol_per_rev = .5*u.mL/u.rev).to(u.rev/u.min)
    >>> round(pump_speed, 6)
    <Quantity(22.105263, 'rev / minute')>
    """

    def __init__(self, Q_sys, C_sys, C_stock):
        """Initialize a reactor of unknown material stock flow rate.

        :param Q_sys: Flow rate of the system
        :type Q_sys: float
        :param C_sys: Concentration of the material in the system
        :type C_sys: float
        :param C_stock: Concentration of the material in the stock
        :type C_stock: float
        """
        self._Q_sys = Q_sys
        self._C_sys = C_sys
        self._C_stock = C_stock

    def Q_sys(self):
        """Return the flow rate of the system.

        :return: Flow rate of the system
        :rtype: float
        """
        return self._Q_sys

    def C_sys(self):
        """Return the concentration of the material in the system.

        :return: Concentration of the material in the system
        :rtype: float
        """
        return self._C_sys

    def C_stock(self):
        """Return the concentration of the material in the stock.

        :return: Concentration of the material in the stock
        :rtype: float
        """
        return self._C_stock

    def Q_stock(self):
        """Return the required flow rate from the stock of material given
        a reactor's desired system flow rate, system concentration, and stock
        concentration.

        :return: Flow rate from the stock of material
        :rtype: float
        """
        return self._Q_sys * (self._C_sys / self._C_stock).to(u.dimensionless)

    @ut.list_handler()
    def rpm(self, vol_per_rev):
        """Return the pump speed required for the reactor's stock of material
        given the volume of fluid output per revolution by the stock's pump.

        :param vol_per_rev: Volume of fluid pumped per revolution (dependent on pump and tubing)
        :type vol_per_rev: float

        :return: Pump speed for the material stock, in revolutions per minute
        :rtype: float
        """
        return Stock.rpm(self, vol_per_rev, self.Q_stock()).to(u.rev/u.min)

    @ut.list_handler()
    def T_stock(self, V_stock):
        """Return the amount of time at which the stock of materal will be
        depleted.

        :param V_stock: Volume of the stock of material
        :type V_stock: float

        :return: Time at which the stock will be depleted
        :rtype: float
        """
        return Stock.T_stock(self, V_stock, self.Q_stock()).to(u.hr)

    @ut.list_handler()
    def M_stock(self, V_stock):
        """Return the mass of undiluted material required for the stock
        concentration.

        :param V_stock: Volume of the stock of material
        :type V_stock: float

        :return: Mass of undiluted stock material
        :rtype: float
        """
        return Stock.M_stock(self, V_stock, self._C_stock)

    @ut.list_handler()
    def V_super_stock(self, V_stock, C_super_stock):
        """Return the volume of super (more concentrated) stock that must be
        diluted for the desired stock volume and stock concentration.

        :param V_stock: Volume of the stock of material
        :type V_stock: float
        :param C_super_stock: Concentration of the super stock
        :type C_super_stock: float

        :return: Volume of super stock to dilute
        :rtype: float
        """
        return Stock.V_super_stock(self, V_stock, self._C_stock, C_super_stock)

    @ut.list_handler()
    def dilution_factor(self, C_super_stock):
        """Return the dilution factor of the concentration of material in the
        stock relative to the super stock.

        :param C_super_stock: Concentration of the super stock
        :type C_super_stock: float

        :return: dilution factor of stock concentration over super stock concentration (< 1)
        :rtype: float
        """
        return Stock.dilution_factor(self, self._C_stock, C_super_stock)



"""
Tests for the research package's tube_sizing module.
"""
import unittest
from aguaclara.core.units import u
import aguaclara.research.stock_qc as stock_qc

C_reactor = stock_qc.Variable_C_Stock(1*u.mL/u.s, 2*u.mg/u.L, 0.4*u.mL/u.s)
Q_reactor = stock_qc.Variable_Q_Stock(4.9*u.mL/u.s, 3.6*u.mg/u.L, 50*u.mg/u.L)


class TestStockQC(unittest.TestCase):

    def assertAlmostEqualQuantity(self, first, second, places=7):
        self.assertAlmostEqual(first.magnitude, second.magnitude, places)
        self.assertAlmostEqual(first.units, second.units, places)

    def test_init(self):
        self.assertEqual(1*u.mL/u.s, C_reactor.Q_sys())
        self.assertEqual(2*u.mg/u.L, C_reactor.C_sys())
        self.assertEqual(0.4*u.mL/u.s, C_reactor.Q_stock())

        self.assertEqual(4.9*u.mL/u.s, Q_reactor.Q_sys())
        self.assertEqual(3.6*u.mg/u.L, Q_reactor.C_sys())
        self.assertEqual(50*u.mg/u.L, Q_reactor.C_stock())

    def test_C_Stock(self):
        self.assertAlmostEqualQuantity(5.0*u.mg/u.L, C_reactor.C_stock())

    def test_Q_Stock(self):
        self.assertAlmostEqualQuantity(0.3528*u.mL/u.s, Q_reactor.Q_stock())

    def test_rpm(self):
        self.assertAlmostEqualQuantity(480*u.rev/u.min, C_reactor.rpm(0.05*u.mL/u.rev))
        self.assertAlmostEqualQuantity(88.2*u.rev/u.min, Q_reactor.rpm(0.24*u.mL/u.rev))

    def test_T_stock(self):
        self.assertAlmostEqualQuantity(3.4722222222222222*u.hr, C_reactor.T_stock(5*u.L))
        self.assertAlmostEqualQuantity(24.722852103804485*u.hr, Q_reactor.T_stock(31.4*u.L))

    def test_M_stock(self):
        self.assertAlmostEqualQuantity(25.0*u.mg, C_reactor.M_stock(5*u.L))
        self.assertAlmostEqualQuantity(1570.0*u.mg, Q_reactor.M_stock(31.4*u.L))

    def test_V_super_stock(self):
        self.assertAlmostEqualQuantity(0.00035714285714285714*u.L,
            C_reactor.V_super_stock(5*u.L, 70*u.g/u.L))
        self.assertAlmostEqualQuantity(0.028035714285714285*u.L,
            Q_reactor.V_super_stock(31.4*u.L, 56*u.g/u.L))

    def test_dilution_factor(self):
        self.assertEqual(7.142857142857142e-05*u.dimensionless, C_reactor.dilution_factor(70*u.g/u.L))
        self.assertEqual(0.0008928571428571429*u.dimensionless, Q_reactor.dilution_factor(56*u.g/u.L))



""" A sedimentation tank of an AguaClara water treatment plant

Example:
    >>> from aguaclara.design.sed_tank import *
    >>> sed_tank = SedimentationTank(q = 60 * u.L / u.s)
    >>> round(sed_tank.diffuser_hl, 5)
    <Quantity(0.00926, 'centimeter')>
"""
from aguaclara.core.units import u
import aguaclara.core.constants as con
import aguaclara.core.materials as mat
import aguaclara.core.pipes as pipe
from aguaclara.core import drills
import aguaclara.core.utility as ut
from aguaclara.design.component import Component
import aguaclara.core.physchem as pc
import aguaclara.core.head_loss as hl

import numpy as np
import math


class SedimentationTank(Component):
    """Design an AguaClara plant's sedimentation tank.

    An sedimentation tank's design relies on the sedimentation channel's design
    in the same plant, but assumed/default values may be used to design an
    sedimentation tank by itself. To design these components in tandem, use
    :class:`aguaclara.design.sed.Sedimentor`.

    Constants:
        - ``INLET_MAN_Q_RATIO (float)``: The ratio of the flow in the inlet
          manifold.
        - ``OUTLET_MAN_HL (float * u.cm)``: The headloss of the outlet manifold
        - ``JET_REVERSER_ND (float * u.inch)``: The nominal diameter of the jet
          reverser.
        - ``JET_PLANE_RATIO (float)``: The ratio for the jet plane
        - ``JET_REVERSER_TO_DIFFUSERS_H (float * u.cm)``: The height between
          the jet reverser and diffusers.
        - ``WALL_THICKNESS (float * u.m)``: The thickness of the sed tank walls
        - ``DIFFUSER_L (float * u.cm)``: The length of a diffuser.

    Design Inputs:
        - ``q (float * u.L / u.s)``: Plant flow rate
          (recommended, defaults to 20L/s)
        - ``temp (float * u.degC)``: Water temperature (recommended, defaults to
          20°C)
        - ``vel_upflow (float * u.mm / u.s)``: Upflow velocity
          (optional, defaults to 1mm/s)
        - ``l_inner (float * u.m)``: The inner length
          (optional, defaults to 5.8m)
        - ``w_inner (float * u.inch)``: The inner width
          (optional, defaults to 42in.)
        - ``diffuser_vel_max (float * u.cm / u.s)``: The max velocity of a
          diffuser (optional, defaults to 44.29 cm/s)
        - ``diffuser_n (int)``:The nunber of diffusers
          (optional, defaults to 108)
        - ``diffuser_wall_thickness (float * u.inch)``: The thickness of the
          wall of a diffuser (optional, defaults to 1.17in.)
        - ``diffuser_sdr (int)``: The standard dimension ratio of a diffuser
          (optional, defaults to 41)
        - ``inlet_man_hl (float * u.cm)``: The headloss of the inlet manifold
          (optional, defaults to 1cm)
        - ``inlet_man_sdr (float)``: The standard dimension ratio of the inlet
          manifold (optional, defaults to 41)
        - ``jet_reverser_sdr (int)``: The standard dimension ratio of the jet
          reverser (optional, defaults to 26)
        - ``plate_settler_angle (float * u.deg)``: The angle of the plate
          settler (optional, defaults to 60°)
        - ``plate_settler_s (float * u.cm)``: Spacing in between plate settlers
          (optional, defaults to 2.5cm)
        - ``plate_settler_thickness (float * u.mm)``: Thickness of a plate
          settler (optional, defaults to 2mm)
        - ``plate_settler_cantilever_l_max (float * u.cm)``: The max length of
          the plate settler cantilever (optional, defaults to 20cm)
        - ``plate_settler_vel_capture (float * u.mm / u.s)``: The capture
          velocity of a plate settler (optional, defaults to 0.12mm/s)
        - ``outlet_man_orifice_hl (float * u.cm)``: The headloss of the
          orifices in the outlet manifold (optional, defaults to 4cm)
        - ``outlet_man_orifice_q_ratio_max (float)``: The max ratio of the flow
          rate for the orifices of the outlet manifold (optional, defaults to 0.8)
        - ``outlet_man_orifice_n_est (int)``: The estimated number of orifices
          for the outlet manifold (optional, defaults to 58)
        - ``outlet_man_sdr (int)``: The standard dimension ratio of the outlet
          manifold (optional, defaults to 41)
        - ``slope_angle (float * u.deg)``: The angle at the bottom of the sed tank
          (optional, defaults to 50°)
        - ``side_slope_to_floc_weir_h_min (float * u.cm)``: The minimum height
          between the side slope and the floc weir. (optional, defaults to 5cm)
        - ``sed_chan_w_outer (float * u.cm)``: The outer width of the
          sedimentation channel (optional, defaults to 60cm)
        - ``sed_chan_weir_thickness (float * u.cm)``: The thickness of the
          sedimentation channel weir (optional, defaults to 5cm)
        - ``floc_weir_to_plate_frame_h (float * u.cm)``: The height from the
          top of the floc weir to the plate settler frame (optional, defaults
          to 10cm)
        - ``hopper_slope_vertical_angle (float * u.deg)``: The angle of the
          hopper wall slopes to vertical (optional, defaults to 60°)

    """
    INLET_MAN_Q_RATIO = 0.8
    OUTLET_MAN_HL = 4. * u.cm
    JET_REVERSER_ND = 3. * u.inch
    JET_PLANE_RATIO = 0.0124
    JET_REVERSER_TO_DIFFUSERS_H = 3.0* u.cm
    WALL_THICKNESS = 0.15 * u.m
    DIFFUSER_L = 15.0 * u.cm

    def __init__(self, **kwargs):
        self.vel_upflow=1.0 * u.mm / u.s
        self.l_inner=5.8 * u.m
        self.w_inner=42.0 * u.inch

        self.diffuser_vel_max=44.29 * u.cm / u.s
        self.diffuser_n=108
        self.diffuser_wall_thickness=1.17 * u.inch
        self.diffuser_sdr=41

        self.inlet_man_hl=1. * u.cm
        self.inlet_man_sdr = 41
        self.jet_reverser_sdr = 26

        self.plate_settler_angle=60.0 * u.deg
        self.plate_settler_s=2.5 * u.cm
        self.plate_settler_thickness=2.0 * u.mm
        self.plate_settler_cantilever_l_max=20.0 * u.cm
        self.plate_settler_vel_capture=0.12 * u.mm / u.s

        self.outlet_man_orifice_hl=4.0 * u.cm
        self.outlet_man_orifice_q_ratio_max=0.8
        self.outlet_man_orifice_n_est = 58
        self.outlet_man_sdr=41

        self.slope_angle=50. * u.deg
        self.side_slope_to_floc_weir_h_min = 5.0 * u.cm
        self.sed_chan_w_outer = 60.0 * u.cm
        self.sed_chan_weir_thickness = 5.0 * u.cm
        self.floc_weir_to_plate_frame_h = 10.0 * u.cm
        self.hopper_slope_vertical_angle = 60.0 * u.deg

        super().__init__(**kwargs)

    @property
    def q_tank(self):
        """The flow rate present in the tank."""
        q_tank = self.l_inner * self.w_inner * self.vel_upflow
        return q_tank.to(u.L / u.s)

    @property
    def diffuser_hl(self):
        """The headloss of the diffuser."""
        return self.inlet_man_hl / self.diffuser_n

    @property
    def diffuser_vel(self):
        """The velocity of the diffuser"""
        diffuser_vel = np.sqrt(2 * con.GRAVITY * self.diffuser_hl)
        return diffuser_vel.to(u.mm / u.s)

    @property
    def diffuser_w_inner(self):
        """The inner width(neglecting walls) of the diffuser."""
        diffuser_w_inner = self.w_inner * self.vel_upflow / self.diffuser_vel
        return diffuser_w_inner.to(u.cm)

    @property
    def diffuser_a(self):
        """The area of the diffuser"""
        diffuser_a = self.q_tank / (self.diffuser_vel * self.diffuser_n)
        return diffuser_a.to(u.cm ** 2)

    @property
    def inlet_man_v_max(self):
        """The maximumum velocity in the inlet manifold."""
        vel_manifold_max = np.sqrt(4 * con.GRAVITY * self.diffuser_hl *
                (1 - self.INLET_MAN_Q_RATIO ** 2) /
                (self.INLET_MAN_Q_RATIO ** 2 + 1)
            )
        return vel_manifold_max.to(u.m / u.s)

    @property
    def inlet_man_nd(self):
        """The nominal diameter of the inlet manifold"""
        diam_inner = np.sqrt(4 * self.q_tank / (np.pi * self.inlet_man_v_max))
        inlet_man_nd = pipe.ND_SDR_available(diam_inner, self.inlet_man_sdr)
        return inlet_man_nd.to(u.cm)

    @property
    def outlet_man_nd(self):
        """The nominal diameter of the outlet manifold."""
        outlet_man_nd = pc.manifold_nd(
            self.q_tank,
            self.OUTLET_MAN_HL,
            self.l_inner,
            self.outlet_man_orifice_q_ratio_max,
            pc.viscosity_kinematic_water(self.temp),
            mat.PVC_PIPE_ROUGH.to(u.m),
            hl.PIPE_EXIT_K_MINOR,
            self.outlet_man_orifice_n_est,
            self.outlet_man_sdr
        )
        return outlet_man_nd

    @property
    def outlet_man_orifice_d(self):
        """The diameter of the orifices in the outlet manifold."""
        Q_orifice = self.q_tank / self.outlet_man_orifice_n_est
        D_orifice = pc.diam_circle(Q_orifice/(con.VC_ORIFICE_RATIO * \
            np.sqrt(2 * con.GRAVITY* self.outlet_man_orifice_hl)))
        return ut.ceil_nearest(D_orifice, drills.DRILL_BITS_D_METRIC)

    @property
    def plate_l(self):
        """The length of a plate in the plate settlers."""
        L_sed_plate = ((self.plate_settler_s * ((self.vel_upflow / \
             self.plate_settler_vel_capture) - 1)
                        + self.plate_settler_thickness * (
                            self.vel_upflow / self.plate_settler_vel_capture))
                     / (np.sin(self.plate_settler_angle) * \
                         np.cos(self.plate_settler_angle))
                     ).to(u.m)
        return L_sed_plate

    @property
    def outlet_man_orifice_q(self):
        """The flow rate in the orifices of the outlet manifold."""
        outlet_man_orifice_q = pc.flow_orifice_vert(
                self.outlet_man_orifice_d,
                self.outlet_man_orifice_hl,
                con.VC_ORIFICE_RATIO
            )
        return outlet_man_orifice_q.to(u.L / u.s)

    @property
    def outlet_man_orifice_spacing(self):
        """The spacing between orifices on the outlet manifold."""
        outlet_man_orifice_spacing = (
            self.l_inner -
            pipe.socket_depth(self.outlet_man_nd) -
            pipe.cap_thickness(self.outlet_man_nd) -
            self.outlet_man_orifice_d
            ) / ((self.q_tank / self.outlet_man_orifice_q) - 1)
        return outlet_man_orifice_spacing

    @property
    def outlet_man_orifice_n(self):
        """The number of orifices on the outlet manifold."""
        outlet_orifice_n = math.floor(
            (
                self.l_inner -
                pipe.socket_depth(self.outlet_man_nd) -
                pipe.cap_thickness(self.outlet_man_nd) -
                self.outlet_man_orifice_d
            ) / self.outlet_man_orifice_spacing
        ) + 1
        return outlet_orifice_n

    @property
    def outlet_orifice_hl(self):
        """The headloss for the orifices of the outlet"""
        outlet_orifice_hl = pc.head_orifice(
            self.outlet_man_nd,
            con.VC_ORIFICE_RATIO,
            self.q_tank / self.outlet_man_orifice_n
            )
        return outlet_orifice_hl.to(u.mm)

    @property
    def side_slopes_w(self):
        """The width of the side slopes."""
        side_slopes_w = (
            self.w_inner -
            pipe.ID_SDR(self.JET_REVERSER_ND, self.jet_reverser_sdr)
            ) / 2
        return side_slopes_w.to(u.m)

    @property
    def side_slopes_h(self):
        """The height of the side slopes."""
        side_slopes_h = np.tan(self.slope_angle) * self.side_slopes_w
        return side_slopes_h.to(u.m)

    @property
    def inlet_man_h(self):
        """The height of the inlet manifold height."""
        inlet_man_h = self.JET_REVERSER_TO_DIFFUSERS_H + self.DIFFUSER_L + \
             ( pipe.OD(self.inlet_man_nd)/ 2 )
        return inlet_man_h

    @property
    def floc_weir_h(self):
        """The height of the floc weir."""
        floc_weir_h = max(
            self.inlet_man_h + (pipe.OD(self.inlet_man_nd) / 2) + \
                mat.CONCRETE_THICKNESS_MIN,
            self.side_slopes_h + self.side_slope_to_floc_weir_h_min
        )
        return floc_weir_h



"""The flocculator of an AguaClara water treatment plant uses turbulence to
cause coagulant and other particles to accumulate, forming flocs.

Example:
    >>> from aguaclara.design.floc import *
    >>> floc = Flocculator(q = 20 * u.L / u.s, hl = 40 * u.cm)
    >>> round(floc.chan_w)
    <Quantity(34, 'centimeter')>
"""
import aguaclara.core.head_loss as hl
import aguaclara.design.human_access as ha
import aguaclara.core.physchem as pc
import aguaclara.core.pipes as pipes
import aguaclara.core.utility as ut
from aguaclara.core.units import u
from aguaclara.design.component import Component
from aguaclara.design.pipeline import Pipe

import numpy as np
from urllib.parse import quote_plus


class Flocculator(Component):
    """Design an AguaClara plant's flocculator.

    A flocculator's design relies on the entrance tank's design in the same
    plant, but assumed/default values may be used to design a flocculator by
    itself. To design these components in tandem, use
    :class:`aguaclara.design.ent_floc.EntTankFloc`.

    Attributes:
        - ``BAFFLE_K (float)``: Minor loss coefficient around a baffle edge
        - ``CHAN_N_MIN (int)``: Minimum channel number
        - ``HS_RATIO_MIN (float)``: Minimum H/S ratio
        - ``HS_RATIO_MAX (float)``: Maximum H/S ratio
        - ``SDR (float)``: Standard dimension ratio
        - ``OBSTACLE_OFFSET (bool)``: Whether the baffle obstacles are offset
          from each other

    Design Inputs:
        - ``q (float * u.L/u.s)``: Flow rate (required)
        - ``temp (float * u.degC)``: Water temperature (optional, defaults to
          20°C)
        - ``ent_l (float * u.m)``: Entrance tank length
          (recommmended, defaults to 1.5m)
        - ``chan_w_max (float * u.inch)``: Maximum width (optional, defaults to
          42")
        - ``l_max (float * u.m)``: Maximum length (optional, defaults to 6m)
        - ``gt (float)``: Collision potential (optional, defaults to 37000)
        - ``hl (float * u.cm)``: Head loss (optional, defaults to 40cm)
        - ``end_water_depth (float * u.m)``: Depth at the end
          (optional, defaults to 2m)
        - ``drain_t (float * u.min)``: Drain time (optional,
          defaults to 30 mins)
        - ``polycarb_sheet_w (float * u.inch)``: Width of polycarbonate sheets
          used to construct baffles (optional, defaults to 42 in)
        - ``sed_chan_inlet_w_pre_weir (float * u.inch)``: Width of the inlet
          sedimentation channel pre-weir (optional, defaults to 42 in)
        - ``dividing_wall_thickness (float * u.cm)``: Thickness of dividing
          walls between each flocculator channel (optional, defaults to 15 cm)
        - ``chan_n_parity (str)``: Parity of the number of channels. Can be
          \'even\', \'odd\', or \'any\' (optional, defaults to \'even\')
    """
    BAFFLE_K = 2.5
    CHAN_N_MIN = 2
    HS_RATIO_MIN = 3.0
    HS_RATIO_MAX = 6.0
    SDR = 41.0 # This is an expert input in ent, should this be an expert
               # input as well? -Oliver L., oal22, 5 Jun 19
    OBSTACLE_OFFSET = True

    _onshape_url = (
        "https://cad.onshape.com/documents/c3a8ce032e33ebe875b9aab4/w/de9ad5474448b34f33fef097/e/08f41d8bdd9a9c90ab396f8a"
    )

    def __init__(self, **kwargs):
        self.ent_l = 1.5 * u.m
        self.chan_w_max = 42.0 * u.inch
        self.l_max = 6.0 * u.m
        self.gt = 37000
        self.hl = 40.0 * u.cm
        self.end_water_depth = 2.0 * u.m
        self.drain_t = 30.0 * u.min
        self.spec = 'sdr41'
        self.drain_pipe = Pipe()
        self.subcomponents = [self.drain_pipe]
        self.polycarb_sheet_w = 42.0 * u.inch
        self.sed_chan_inlet_w_pre_weir = 42.0 * u.inch
        self.dividing_wall_thickness = 15.0 * u.cm
        self.chan_n_parity = 'even'

        super().__init__(**kwargs)

        self._set_drain_pipe()
        super().set_subcomponents()

        if self.chan_n_parity not in ('even', 'odd', 'any'):
            raise AttributeError(
                'chan_n_parity must be set to \'even\', \'odd\', or \'any\'.'
            )

    @property
    def vel_grad_avg(self):
        """The average velocity gradient of water."""
        vel_grad_avg = ((u.standard_gravity * self.hl) /
                        (pc.viscosity_kinematic_water(self.temp) * self.gt)).to(u.s ** -1)
        return vel_grad_avg

    @property
    def retention_time(self):
        """The hydraulic retention time neglecting the volume
        created by head loss.
        """
        retention_time = (self.gt / self.vel_grad_avg).to(u.s)
        return retention_time

    @property
    def vol(self):
        """The target volume (not counting the volume added by head loss)."""
        return (self.q * self.retention_time).to(u.m ** 3)

    @property
    def chan_w_min_hs_ratio(self):
        """The minimum channel width."""
        chan_w_min_hs_ratio = (
            (self.HS_RATIO_MIN * self.q / self.end_water_depth) *
            (
                self.BAFFLE_K / (
                    2 * self.end_water_depth *
                    pc.viscosity_kinematic_water(self.temp) *
                    self.vel_grad_avg ** 2
                )
            ) ** (1/3)
        ).to(u.cm)
        return chan_w_min_hs_ratio

    @property
    def chan_w_min(self):
        """The minimum channel width."""
        return ut.min(self.chan_w_min_hs_ratio, self.polycarb_sheet_w).to(u.cm)

    @property
    def chan_n(self):
        """The minimum number of channels based on the maximum
        possible channel width and the maximum length of the channels.
        """
        if self.q < 16 * u.L / u.s:
            return 1
        else:
            chan_n = ((
                (self.vol /
                 (self.polycarb_sheet_w * self.end_water_depth)
                 ) + self.ent_l
            ) / self.chan_l).to_base_units()

            if self.chan_n_parity is 'even':
                return ut.ceil_step(chan_n, step=2)
            elif self.chan_n_parity is 'odd':
                return ut.ceil_step(chan_n, step=2) - 1
            elif self.chan_n_parity is 'any':
                return ut.ceil_step(chan_n, step=1)

    @property
    def chan_w_min_gt(self):
        """The channel width minimum regarding the collision potential."""
        chan_w_min_gt = self.vol / (
            self.end_water_depth * (self.chan_n * self.chan_l - self.ent_l)
        )
        return chan_w_min_gt.to(u.cm)

    @property
    def chan_w(self):
        """The channel width."""
        chan_w = ut.ceil_step(
            ut.max(self.chan_w_min_gt, self.chan_w_min),
            step=1 * u.cm
        )
        return chan_w

    @property
    def l_max_vol(self):
        """The maximum length depeneding on the volume."""
        l_max_vol = self.vol / \
            (self.CHAN_N_MIN * self.chan_w_min * self.end_water_depth)
        return l_max_vol.to(u.m)

    @property
    def chan_l(self):
        """The channel length."""
        chan_l = min(self.l_max, self.l_max_vol)
        return chan_l.to_base_units()

    @property
    def expansion_h_max(self):
        """"The maximum distance between expansions for the largest
        allowable H/S ratio.
        """
        expansion_h_max = (
            (
                (self.BAFFLE_K /
                        (
                            2 * pc.viscosity_kinematic_water(self.temp) *
                            (self.vel_grad_avg ** 2)
                        )
                    ) *
                    (self.q * self.HS_RATIO_MAX / self.chan_w) ** 3
                ) ** (1/4)
            ).to(u.m)
        return expansion_h_max

    @property
    def expansion_n(self):
        """The minimum number of expansions per baffle space."""
        return np.ceil(self.end_water_depth / self.expansion_h_max)

    @property
    def expansion_h(self):
        """The height between flow expansions."""
        return (self.end_water_depth / self.expansion_n).to(u.cm)

    @property
    def baffle_s(self):
        """The spacing between baffles."""
        baffle_s = (
            (self.BAFFLE_K /
             (
                 (2 * self.expansion_h * (self.vel_grad_avg ** 2) *
                  pc.viscosity_kinematic_water(self.temp))
             ).to_base_units()
                ) ** (1/3) *
                self.q / self.chan_w
            ).to(u.cm)
        return baffle_s

    @property
    def obstacle_n(self):
        """The number of obstacles per baffle."""
        return self.end_water_depth / self.expansion_h - 1

    @property
    def contraction_s(self):
        """The space in the baffle by which the flow contracts."""
        return self.baffle_s * 0.6

    @property
    def obstacle_pipe_od(self):
        """The outer diameter of an obstacle pipe. If the available pipe is
        greater than 1.5 inches, the obstacle offset will become false."""
        pipe_od = pipes.od_available(self.contraction_s)

        if pipe_od > 1.5 * u.inch:
            self.OBSTACLE_OFFSET = False
            pipe_od = pipes.od_available(pipe_od / 2)

        return pipe_od

    def _set_drain_pipe(self):
        drain_k_minor = \
            hl.PIPE_ENTRANCE_K_MINOR + \
            hl.PIPE_ENTRANCE_K_MINOR + \
            hl.PIPE_EXIT_K_MINOR

        chan_pair_a = 2 * self.chan_l * self.chan_w
        drain_id = (
            np.sqrt(8 * chan_pair_a / (np.pi * self.drain_t) *
                    np.sqrt(
                self.end_water_depth * drain_k_minor /
                (2 * u.standard_gravity)
            )
            )
        ).to_base_units()

        self.drain_pipe = Pipe(
            id=drain_id,
            k_minor=drain_k_minor,
            spec=self.spec
        )

    @property
    def onshape_url_configured(self):
        # Make the configuration string for the flocculator concrete. {{ and }}
        # are used so that str.format() doesn't recognize them.
        concrete_config = (
            '{{"w_channel":"{}", "h_channel":"{}", "l_channel":"{}", "s_baffle"'
            ':"{}", "n_channel":{}, "t_wall":"{}"}}'.format(
                self.chan_w,
                self.end_water_depth,
                self.chan_l,
                self.baffle_s,
                self.chan_n,
                self.dividing_wall_thickness
            )
        )
        # concrete_config = quote_plus(concrete_config)
        encoded_config = '?configuration='

        encoded_config += quote_plus('Concrete_config=' + concrete_config + ';')
        encoded_config += quote_plus('Channel_L=' + str(self.chan_l) + ';')

        configured_url = self._onshape_url + encoded_config
        return configured_url



"""Design abilities for AguaClara drinking water treatment plant components

This module allows plant component classes to

#. specify expert inputs and their defaults automatically
#. propogate plant-wide design inputs throughout all of its subcomponents
#. reconfigure Onshape Part Studios and Assemblies that represent plant
   components.

If a ``Component`` is instantiated with custom subcomponents, and the
subcomponents have their own ``q`` and/or ``temp`` specified:

.. code-block:: python

    plant = Plant(
            q = 40 * u.L / u.s,
            filter = Filter(q = 20 * u.L / u.s)
        )

then the subcomponets will preserve their ``q`` and ``temp``:

.. code-block:: python

    >>> plant.q
    40 liter / second
    >>> plant.filter.q
    20 liter / second

Otherwise, subcomponents will inherit the main component's ``q`` and ``temp``:

.. code-block:: python

    >>> plant = Plant(q = 40 * u.L / u.s)
    >>> plant.q
    40 liter / second
    >>> plant.filter.q
    40 liter / second

.. # TODO: update the a code example with the complete Onshape design flow.
"""
from aguaclara.core.units import u
import aguaclara.core.utility as ut

import numpy as np
import json
from pprint import pprint
from abc import ABC, abstractmethod
from urllib.parse import quote_plus


class Component(ABC):
    """An abstract class representing AguaClara plant components.
    
    This class provides subclasses with the ability to record and propogate a
    configuration of plant design variables (``q`` and ``temp``) for a component
    and all of its subcomponents.
    """
    Q_DEFAULT = 20 * u.L / u.s
    TEMP_DEFAULT = 20 * u.degC
    onshape_url_default = ''

    def __init__(self, **kwargs):
        self.q = self.Q_DEFAULT
        self.temp = self.TEMP_DEFAULT

        # Update the Component with new expert inputs, if any were given
        self.__dict__.update(**kwargs)

    def set_subcomponents(self):
        """Set the plant-wide inputs of all subcomponents.

        Call this function at the end of a subclass's ``__init__()`` set ``q``
        and ``temp`` for subcomponents, except when a subcomponent specifies its
        own custom ``q`` and ``temp``.
        """
        for subcomp in self.subcomponents:
            # Set the subcomponent's ``q`` and ``temp`` to match this component
            # unless they were changed during instantiation.
            if subcomp.q == self.Q_DEFAULT:
                subcomp.q = self.q
            if subcomp.temp == self.TEMP_DEFAULT:
                subcomp.temp = self.temp

            # Recursively set sub-subcomponents.
            if hasattr(subcomp, 'subcomponents'):
                subcomp.set_subcomponents()
        
    def serialize_properties(self):
        """Return the properties (fields and ``@property`` functions) of a 
        component as a dictionary string.
        """
        properties = {}
        ignored_properties = [
            '__dict__',
            '__doc__',
            '__module__',
            '__weakref__',
            'subcomponents',
            'Q_DEFAULT',
            'TEMP_DEFAULT',
            '__abstractmethods__',
            '_abc_cache',
            '_abc_negative_cache',
            '_abc_negative_cache_version',
            '_abc_registry',
            'onshape_config'
        ]
        # Get all of the object's fields
        for var_name in dir(self):
            value = getattr(self, var_name)

            # Serialize subcomponents to strings so that they are accessible by
            # Onshape's Super Derive feature
            if isinstance(value, Component):
                properties[var_name] = str(value.serialize_properties())

            # Serialize non-component properties
            elif not callable(value) and var_name not in ignored_properties:

                # Serialize lists or arrays
                try: 
                    if type(value.magnitude) is np.ndarray:
                        properties[var_name] = ut.array_qtys_to_strs(value)
                    else:
                        properties[var_name] = str(value)
                
                # Serialize non-iterable properties
                except AttributeError: 
                    properties[var_name] = str(value)
    
        return properties

    def print_properties(self):
        """Print the serialized properties with pretty indentation."""
        pprint(self.serialize_properties())
    
    def write_properties_to_file(self):
        """Append the properties of a component to a file. If it does not exist,
        then the file is created.
        
        Args:
            - ``filename (str)``: The name of the file
        """
        filename = "props.json"
        json.dump(self.serialize_properties(), open(filename, mode='w'),
            indent = 4)
        print("Properties of component can be found in file props.json")

    def _encode_onshape_config(self, config):
        encoding = ''
        for key, value in onshape_config.items():
            if type(value) is u.Quantity:
                value = str(value)
            elif type(value) is dict:
                value = _encode_onshape_config(value)
            encoding += key

    @property
    def onshape_url_configured(self):
        onshape_url_configured = (
            self.onshape_url_default + 
            '?configuration=' +
            self._encode_onshape_config(self.onshape_config)
        )
        return onshape_url_configured

# TODO: get the following to work in the general case when passing a simple
# nested dictionary to Onshape.

def nested_dict_to_str(dict_, level=0):
    for key, value in dict_.items():
        if type(value) is u.Quantity:
            dict_[key] = str(value)
        elif type(value) is dict:
            dict_[key] = nested_dict_to_str(value, level - 1)
    
    if level <= 0:
        dict_ = str(dict_)
    return dict_

def encode_onshape_config(config):
    encoding = ''
    for key, value in config.items():
        if type(value) in [u.Quantity, dict]:
            value = str(value)
        value = quote_plus(value)
        encoding += '{}={};'.format(key, value)

    return encoding

def onshape_url_configured(base_url, encoding):
    onshape_url_configured = (
        base_url + 
        '?configuration=' +
        quote_plus(encoding)
    )
    return onshape_url_configured